/* nurbs.c - Introduction  la manipulation des NURBS avec OpenGL */

/* Copyright (C) 2001 Jean-Max Redonnet
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>
#include <GL/glut.h>
#include <GL/glu.h>

#define WINWIDTH 600
#define WINHEIGHT 600

/* nombre de surfaces  traiter */
#define NB_SURF 3

/*
 * paramtre dfinissant l'impact de l'alatoire
 * 0:aucun impact : on dessine des rectangle... passionnant :-( 
 * 5 impact mesur
 * 10:trs gros impact
 */
/*#define ALEA 0*/

/*
 * Les poles de la surface sont stocks squentiellement dans le pointeur
 * ppoles. Ces macros permettent de les retrouver  partir de leur indice en u
 * et en v.
 */ 
#define PPOLE(l,c)	(ppole[l*npv + c])
#define PSURF_PT(l,c)	(psurf->ppole[l*psurf->npv+c])

typedef struct Vec4 {
	GLfloat x;
	GLfloat y;
	GLfloat z;
	GLfloat h;
	}Vec4;
	
typedef struct Surf *psuiv;

typedef struct Surf {
	int id_surf;
	int	npu;
	int	npv;
	int dgu;
	int dgv;
	int	nku;
	int	nkv;
	float *knotsu;
	float *knotsv;
	struct Vec4 *ppole;
	struct Surf *psuiv;
	}Surf;

typedef struct Visu {
	GLfloat *light_ambient;
	GLfloat *surf_ambient;
	GLfloat *light_position;
	int ctrl_pts;
	GLint rendu;
	}Visu;

typedef struct Model {
	Surf *psurf0;
	Visu *pvisu;
	GLUnurbsObj *NurbsModel;
	GLvoid *font_style;
	int alea;
	}Model;

/* une petite numration pour faciliter la lecture */
enum{NON, OUI};

Model *pmodel;

/* 
 * Dfinition des squences nodales (nku et nkv termes)
 * Attention, on doit avoir nku = npu + dgu + 1 et nkv = npv + dgv + 1
 */
GLfloat knotsu[8]={0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0};
GLfloat knotsv[6]={0.0, 0.0, 0.0, 1.0, 1.0, 1.0};

GLfloat light_ambient[4]={0.1, 0.1, 0.1, 1.0};
/*GLfloat surf_ambient[4]={0.4, 0.4, 0.7, 1.0};*/
GLfloat light_position[4]={1.0, 1.0, 1.0, 0.0};

/********************************************************************/
/*                                                                  */
/* AFFICHAGES                                                       */
/*                                                                  */
/********************************************************************/

/*
 * Affichage d'une chaine de caractres
 */
void drawstr(GLfloat x, GLfloat y, GLfloat z, char* format, ...)
{
	va_list args;
	char buffer[255], *s;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	glRasterPos3f(x, y, z);
	for (s = buffer; *s; s++)
	glutBitmapCharacter(pmodel->font_style, *s);
}

/*
 * Affichage d'un repre
 */
void plot_rep(char *labX, char *labY, char *labZ, float size)
{
	glLineWidth(2.0);
	glBegin(GL_LINES);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(size, 0.0, 0.0);
	glEnd();
	drawstr( size + size/10, 0.0, 0.0, labX );
	glBegin(GL_LINES);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, size, 0.0);
	glEnd();
	drawstr( 0.0, size + size/10, 0.0, labY );
	glBegin(GL_LINES);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(0.0, 0.0, size);
	glEnd();
	drawstr( 0.0, 0.0, size+size/10, labZ );
	glLineWidth(1.0);
}

/*
 * Affichage du rseau caractristique         
 * (le rseau caractristique est le rseau qui relie tous les ples entre eux) 
 */
void plot_res(Surf * psurf0)
{
	int i, j, k;
	Surf * psurf;
	
	glPointSize(5.0);
	glDisable(GL_LIGHTING);
	glColor3f(1.0, 1.0, 0.0);

	psurf = psurf0;
	for(k=0 ; k<NB_SURF ; k++){
		/* Affichage des points de controle (ples) */
		glBegin(GL_POINTS);
		for(i=0;i < psurf->npu;i++) {
			for(j=0;j < psurf->npv;j++) {
				glVertex3f(PSURF_PT(i,j).x, PSURF_PT(i,j).y, PSURF_PT(i,j).z);
			}
		}
		glEnd();

		/* Affichage des lignes entre les ples */
		for(i=0; i < psurf->npu;i++) {
			glBegin(GL_LINE_STRIP);
			for (j=0; j < psurf->npv; j++) {
				glVertex3f(PSURF_PT(i,j).x, PSURF_PT(i,j).y, PSURF_PT(i,j).z);
			}
			glEnd();
		}
	
		for(j=0;j < psurf->npv;j++) {
			glBegin(GL_LINE_STRIP);
			for (i=0; i < psurf->npu; i++) {
				glVertex3f(PSURF_PT(i,j).x, PSURF_PT(i,j).y, PSURF_PT(i,j).z);
			}
			glEnd();	
		}
	psurf=psurf->psuiv;
	}
	glEnable(GL_LIGHTING);
}

/*
 * Affichage d'une surface
 */
void drawSurf(Model * pmodel)
{
	int i, j;
	Surf *psurf;
	GLUnurbsObj *NurbsModel;
	GLfloat surf_ambient[4];
	
	psurf = pmodel->psurf0;
	NurbsModel = pmodel->NurbsModel;
	gluNurbsProperty(NurbsModel, GLU_DISPLAY_MODE, pmodel->pvisu->rendu);
	for(i=0 ; i<NB_SURF ; i++){
		/*
		 * En plaant le gnrateur de couleurs ici, il suffit de cliquer sur
		 * le bouton gauche de la souris pour changer de couleurs car
		 * glutPostRedisplay() est appel automatiquement  la sortie du menu
		 */
		for (j=0 ; j<4 ; j++){
			surf_ambient[j] = rand()/(RAND_MAX+1.0);
		}
		glMaterialfv(GL_FRONT_AND_BACK,
				 	 GL_AMBIENT_AND_DIFFUSE, surf_ambient);

		gluBeginSurface(NurbsModel);
		gluNurbsSurface(NurbsModel,
						(GLint) psurf->nku,
						(GLfloat *) psurf->knotsu,
						(GLint) psurf->nkv,
						(GLfloat *) psurf->knotsv,
						(GLint) psurf->npv * 4,
						4,
						(GLfloat *) psurf->ppole,
						(GLint) psurf->dgu + 1,
						(GLint) psurf->dgv + 1,
						GL_MAP2_VERTEX_4);
		gluEndSurface(NurbsModel);
		psurf = psurf->psuiv;
	}
}

/********************************************************************/
/*                                                                  */
/* CALLBACKS                                                        */
/*                                                                  */
/********************************************************************/
void Idle(void)
{
	/*
	 * Dcommentez la ligne suivante pour obtenir un effet plus "flashee"
	 */
	 
	/*glutPostRedisplay();*/
}

void Reshape(int width, int height)
{
	glViewport( 0, 0, width, height );
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

void Key( unsigned char key, int x, int y )
{
	switch (key) {
		case 27:
			exit(0);
			break;
	}
	glutPostRedisplay();
}


void Display(void)
{
	char *s = "Double Buffer";
	float t; /* taille de la scene */
	
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	t = 3;
	glOrtho(-t, t, -t, t, 0, 2*t);
	gluLookAt(t*sqrt(3.0)/3.0, t*sqrt(3.0)/3.0, t*sqrt(3.0)/3.0,
	          0, 0, 0, 0, 1, 0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* Affichage du repre */
	if(glIsEnabled(GL_LIGHTING)) glDisable(GL_LIGHTING);
	glColor3f(1.0, 1.0, 1.0);
	plot_rep("X", "Y", "Z", 0.2);

	/* Affichage des surfaces	*/
	glColor3f(1.0, 0.0, 0.0);
	glEnable(GL_LIGHTING);
	drawSurf(pmodel);
	
	/* Affichage du polygone de contrle (rseau caractristique) */
	if(pmodel->pvisu->ctrl_pts == OUI) {
		plot_res(pmodel->psurf0);
	}

	/* Affichage des annotations fixes */   
	if(glIsEnabled(GL_LIGHTING)) glDisable(GL_LIGHTING);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(-1, 1, -1, 1, 0, 1);
	gluLookAt(0,0,1,0,0,0,0,1,0);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glColor3f( 1.0, 1.0, 1.0 );
	drawstr(-0.95, -0.95, 1, s );

	glutSwapBuffers();
}

/********************************************************************/
/*                                                                  */
/* INITIALISATIONS                                                  */
/*                                                                  */
/********************************************************************/
float delta_xy(){

  return (float)( pmodel->alea/2 * (rand()/(RAND_MAX+1.0) - 0.5));
}

/****************/
/* Surface init */
/****************/
Surf *chg_surf(int id)
{
	/* Charge une surface */
	int i,j;

	int npu=4, npv=3, dgu=3, dgv=2;
	int nku, nkv;
	Vec4 *ppole;
	Surf *psurf;
	
	srand(time(NULL)+id);
	psurf = (Surf *) malloc(sizeof(Surf));
	psurf->id_surf=id;
	psurf->npu=npu;
	psurf->npv=npv;
	psurf->dgu=dgu;
	psurf->dgv=dgv;
	nku = npu + dgu + 1;
	nkv = npv + dgv + 1;
	psurf->nku=nku;
	psurf->nkv=nkv;
	psurf->knotsu = (GLfloat *)malloc((nku + 1)*sizeof(GLfloat));
	psurf->knotsu = knotsu;
	psurf->knotsv = (GLfloat *)malloc((nkv + 1)*sizeof(GLfloat));
	psurf->knotsv = knotsv;

	ppole = (Vec4 *) malloc(npv * npu * sizeof(Vec4));
	
	for (j=0;j<npv;j++){
		for (i=0;i<npu;i++){
			PPOLE(i,j).x = i - npu/2+0.5 + delta_xy();
			PPOLE(i,j).y = j - npv/2 + delta_xy();
			PPOLE(i,j).z = (float)(id - NB_SURF + 1
								   + 10*pmodel->alea * rand()/(RAND_MAX+1.0));
			PPOLE(i,j).h = 1.0; /* on met tous les poids  1 */
		}
	}
	psurf->ppole = ppole;
	
	return psurf;
}

void init_surface()
{
	int i;
	Surf * psurf;
	
	printf("Chargement des surfaces \n");
	psurf = (Surf *)malloc(sizeof(Surf));
	pmodel->psurf0 = chg_surf(1);
	psurf = pmodel->psurf0;
	printf("\t surface 1 charge \n");

	for (i=2; i<=NB_SURF; i++) {
		psurf->psuiv = chg_surf(i);
		printf("\t surface %d charge \n", i);
		psurf = psurf->psuiv;
	}
	printf("%i surface(s) charge(s) \n", i-1);
}

void Init()
{
 	/* Dfinition de l'clairage */
	pmodel->pvisu->light_ambient = (GLfloat *)malloc(4.0*sizeof(GLfloat));
	pmodel->pvisu->surf_ambient = (GLfloat *)malloc(4.0*sizeof(GLfloat));
	pmodel->pvisu->light_position = (GLfloat *)malloc(4.0*sizeof(GLfloat));
	pmodel->pvisu->light_ambient=light_ambient;
/*	pmodel->pvisu->surf_ambient=surf_ambient;*/
	pmodel->pvisu->light_position=light_position;
	
	glLightfv(GL_LIGHT0, GL_AMBIENT, pmodel->pvisu->light_ambient);
	glLightfv(GL_LIGHT0, GL_POSITION, pmodel->pvisu->light_position);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
/*	glMaterialfv(GL_FRONT_AND_BACK,
				 GL_AMBIENT_AND_DIFFUSE, pmodel->pvisu->surf_ambient);
*/	glEnable(GL_LIGHT0);
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);
	
	pmodel->pvisu->ctrl_pts = NON;
	pmodel->pvisu->rendu = GLU_FILL;
	
	/* Initialisation des surfaces */
	init_surface();
	
	pmodel->NurbsModel = gluNewNurbsRenderer();
	gluNurbsProperty(pmodel->NurbsModel, GLU_SAMPLING_TOLERANCE, 25.0);
	gluNurbsProperty(pmodel->NurbsModel,
	                 GLU_DISPLAY_MODE, pmodel->pvisu->rendu);
	glEnable(GL_DEPTH_TEST);
}


/********************************************************************/
/*                                                                  */
/* MENUS                                                            */
/*                                                                  */
/********************************************************************/

void Menu(int value)
{
	switch (value) {
	case 0:
		exit(0);
		break;
	case 1:
		pmodel->pvisu->ctrl_pts = OUI;
		break;
	case 2:
		pmodel->pvisu->ctrl_pts = NON;
		break;
	case 3:
		pmodel->pvisu->rendu = GLU_FILL;
		break;
	case 4:
		pmodel->pvisu->rendu = GLU_OUTLINE_POLYGON;
		break;
	case 5:
		init_surface();
		break;
	}
	glutPostRedisplay();
}

void SubMenu(int value)
{
	pmodel->alea = value;
	init_surface();
	glutPostRedisplay();
}

void creer_menu(void)
{
	int i, m;
	char *s;
	
	s = (char *)malloc(sizeof(char));
	m = glutCreateMenu(SubMenu);
	for (i=0; i<=10 ; i++){
		sprintf(s, "%d", i);
		glutAddMenuEntry(s, i);
	}
	free((void *) s);
		
	glutCreateMenu(Menu);
	glutAddMenuEntry("Voir rseau caractristique", 1);
	glutAddMenuEntry("Cacher rseau caractristique", 2);
	glutAddMenuEntry("Rendu raliste", 3);
	glutAddMenuEntry("Rendu filaire", 4);
	glutAddMenuEntry("Rgnerer", 5);
	glutAddSubMenu("Alea", m);
	glutAddMenuEntry("Quitter", 0);
	glutAttachMenu(GLUT_RIGHT_BUTTON);

}

/********************************************************************/
/*                                                                  */
/* MAIN                                                             */
/*                                                                  */
/********************************************************************/
int main( int argc, char *argv[] )
{	
	pmodel = (Model *)malloc(sizeof(Model));
	pmodel->psurf0 = (Surf *)malloc(sizeof(Surf));
	pmodel->pvisu = (Visu *)malloc(sizeof(Visu));
	pmodel->font_style = GLUT_BITMAP_HELVETICA_12;
	pmodel->alea = 5;
		
	glutInit( &argc, argv );
	glutInitWindowSize( WINWIDTH, WINHEIGHT );

	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
	glutCreateWindow("Dmo Nurbs");

	Init();
	creer_menu();
	
	glutReshapeFunc(Reshape);
	glutDisplayFunc(Display);
	glutKeyboardFunc(Key);
	glutIdleFunc(Idle);
	
	glutMainLoop();

	free((void *) pmodel->psurf0);
	free((void *) pmodel->pvisu);
	free((void *) pmodel);

	return 0;
}
